from avatar_utils.validation.schemas import (SuccessResponseSchema,
                                             ErrorResponseSchema,
                                             RootResponseSchema,
                                             ResponseSchema,
                                             ObjectsPostSchema,
                                             ServicePostRequestSchema,
                                             ServiceResultPostResponseSchema,
                                             ServicePostResponseSchema)
from avatar_utils.validation.validate_json import validate_json

__all__ = ['validate_json',
           'ResponseSchema',
           'SuccessResponseSchema',
           'ErrorResponseSchema',
           'RootResponseSchema',
           'ObjectsPostSchema',
           'ServicePostRequestSchema',
           'ServiceResultPostResponseSchema',
           'ServicePostResponseSchema']



from logging import getLogger
from typing import Optional

import jwt
from jwt.exceptions import PyJWTError, InvalidTokenError
from keycloak import KeycloakOpenID

from avatar_utils.sso_helper.constants import (
    ITMO_SSO_SERVER,
    ITMO_REALM,
    USER_TOKEN_AUDIENCE,
    PUBLIC_KEY_WRAPPER,
)
from avatar_utils.sso_helper.flask_auth_header import FlaskAuthHeader

logger = getLogger()


class SSOHelper:

    def __init__(self,
                 client_id: str,
                 client_secret: str,
                 server_url: str = ITMO_SSO_SERVER,
                 verify_token: bool = True):

        self.verify_token = verify_token
        self.auth_header = FlaskAuthHeader()
        self.keycloak_client: KeycloakOpenID = KeycloakOpenID(server_url=server_url,
                                                              client_id=client_id,
                                                              realm_name=ITMO_REALM,
                                                              client_secret_key=client_secret)
        self._public_key: Optional[str] = None

    @property
    def public_key(self) -> str:
        if not self._public_key:
            self._public_key = self.keycloak_client.public_key()
            logger.info(f'RECEIVED PUBLIC KEY:\n{self._public_key}')
        return PUBLIC_KEY_WRAPPER.format(public_key=self._public_key)

    def extract_payload(self,
                        token: str,
                        verify: Optional[bool] = None,
                        raise_exc: bool = False) -> Optional[dict]:

        if verify is None:
            verify = self.verify_token

        if self.verify_token and not self.public_key:
            if raise_exc:
                raise ValueError('***** PUBLIC KEY IS ABSENT *****')
            logger.error('***** PUBLIC KEY IS ABSENT *****')
            return

        try:
            return jwt.decode(token,
                              key=self.public_key,
                              audience=USER_TOKEN_AUDIENCE,
                              verify=verify)
        except (PyJWTError, InvalidTokenError, ValueError) as err:
            logger.warning(f'{err.__class__.__name__}: {token}')
            if raise_exc:
                raise err
            return

    @staticmethod
    def make_token(sso_user_id: str,
                   key: str = 'secret',
                   **payload_fields) -> str:

        payload_fields.update(dict(sub=sso_user_id,
                                   aud=USER_TOKEN_AUDIENCE))
        token = jwt.encode(payload=payload_fields, key=key)
        return token.decode()



from dataclasses import dataclass as python_dataclass, field

from marshmallow_dataclass import dataclass

from avatar_utils.objects.abstracts.abstract_object import AbstractObject


@dataclass
@python_dataclass
class Image(AbstractObject):
    url: str = None

    repr_type: str = None

    id: int = None
    stars: int = None



from logging import getLogger
from typing import Optional

from flasgger import Swagger
from flask import Flask
from marshmallow import ValidationError
from requests import post, ConnectionError

from avatar_utils.validation import SuccessResponseSchema
from avatar_utils.validation.constants import ServiceHTTPCodes

logger = getLogger(__name__)


def service_registration(swagger: Swagger,
                         url: str,
                         username: str,
                         password: str,
                         service_name: Optional[str] = None,
                         service_url: Optional[str] = None) -> Optional[dict]:

    app: Flask = swagger.app
    registration_response: Optional[dict] = None

    payload = dict(username=username, password=password, name=service_name, url=service_url)

    with app.app_context():
        spec = swagger.get_apispecs()
        logger.debug(f'spec = {spec}')
        payload.update(dict(api_description=spec))

    payload = dict((k, v) for k, v in payload.items() if v is not None)

    try:
        response = post(url=url, json=payload)

        logger.debug(f'{response.status_code} {response.reason}: {response.text}')
        if response.status_code == ServiceHTTPCodes.OK.value:
            registration_response = SuccessResponseSchema().loads(response.text)
    except ValidationError as err:
        logger.error(f'Incorrect response from server: {err.messages}')
    except TypeError as err:
        logger.error(f'Dump specification to JSON error: {err}')
    except ConnectionError as err:
        logger.error(f'Registration host unavailable: {err}')
    except Exception as err:  # noqa W0703
        logger.error(f'Unknown error ({err.__class__.__name__}) during registration: {err}')

    logger.info(f'Registration state: {"FAIL" if registration_response is None else "PASS"}')
    return registration_response



from dataclasses import dataclass, field
from json import dumps

from aiohttp import ClientTimeout
from sys import getsizeof
from typing import Optional

from marshmallow import Schema, fields


class RequestDataSchema(Schema):
    url = fields.Str(required=True)
    method = fields.Str(required=False, default='GET')
    count = fields.Int(required=False, default=1)
    timeout = fields.Int(required=False, allow_none=True, default=None)
    headers = fields.Dict(required=False, default=None)
    json = fields.Dict(required=False, default=None)


@dataclass
class RequestData:

    url: str
    method: str = 'GET'
    count: int = 1
    timeout: Optional[ClientTimeout] = None
    headers: Optional[dict] = None
    json: Optional[dict] = None

    size: int = field(init=False)

    def __post_init__(self):
        self.size = 0
        if self.json is not None:
            self.size = getsizeof(dumps(self.json))

    @staticmethod
    def make(request_data):
        req_schema = RequestDataSchema()

        if isinstance(request_data, str):
            request_data = dict(url=request_data)

        if isinstance(request_data, dict):
            req_schema.load(request_data)
            request_data = req_schema.dump(request_data)

        if not isinstance(request_data, RequestData):
            seconds = request_data.get('timeout')
            timeout = ClientTimeout(total=0 if seconds is None else seconds)
            return RequestData(method=request_data['method'],
                               url=request_data['url'],
                               count=request_data['count'],
                               json=request_data['json'],
                               timeout=timeout)
        return request_data



from dataclasses import dataclass as python_dataclass, field
from typing import List, Any

from marshmallow_dataclass import dataclass

from avatar_utils.objects.abstracts.serializable import Serializable


@dataclass
@python_dataclass
class FeedItem(Serializable):
    objects: List[Any] = field(default_factory=list)

    id: int = None

    service: str = None
    service_header: str = None
    route: str = None

    pinned: bool = None
    score: float = None
    date: str = None

    pop_up: bool = None
    url: str = None

    label: str = None

    repr_type: str = None



from app import db


class CrawlerTask(db.Model):
    __tablename__ = 'crawler_task'

    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    task_id = db.Column(db.BigInteger, unique=True, index=True)
    name = db.Column(db.String, index=True)
    last_crawled = db.Column(db.BigInteger, default=0)
    task_body = db.Column(db.Text)

    def __repr__(self):
        return "<{}:{}:{}>".format(self.task_id, self.name, self.last_crawled)



from dataclasses import dataclass as python_dataclass, field
from typing import List, Any

from marshmallow_dataclass import dataclass

from avatar_utils.objects import Image
from avatar_utils.objects.abstracts.abstract_object import AbstractObject


@dataclass
@python_dataclass
class Person(AbstractObject):
    tags: List[Any] = field(default_factory=list)
    lastname: str = None
    firstname: str = None
    middlename: str = None
    image: Image = None

    repr_type: str = None

    id: int = None
    stars: int = None



from logging import getLogger

from marshmallow import Schema, fields
from marshmallow.validate import Equal

logger = getLogger('sample-service')


class ResponseSchema(Schema):
    headers = fields.Dict(required=True, allow_none=False, default=dict())
    message = fields.Str(required=True, allow_none=True, default='')
    result = fields.Dict(required=True, allow_none=True)
    success = fields.Bool(required=True, allow_none=False)


class RootResponseSchema(Schema):
    name = fields.Str(required=True, allow_none=False)
    version = fields.Str(required=False, allow_none=True, default=None)


class SuccessResponseSchema(ResponseSchema):
    success = fields.Bool(required=True, allow_none=False, validate=Equal(True), default=True)


class ErrorResponseSchema(ResponseSchema):
    success = fields.Bool(required=True, allow_none=False, validate=Equal(False), default=False)


class ObjectsPostSchema(Schema):
    objects = fields.List(fields.Raw(), required=False)


class ServicePostRequestSchema(ObjectsPostSchema):
    pass


class ServiceResultPostResponseSchema(ObjectsPostSchema):
    label = fields.Str(required=False)
    objects = fields.List(fields.Raw(), required=True)


class ServicePostResponseSchema(SuccessResponseSchema):
    result = fields.Nested(ServiceResultPostResponseSchema, required=True)



class Env:
    protocol: str
    host: str
    port: int = 8000

    @property
    def base_url(self):
        return f'{self.protocol}{self.host}:{self.port}'


class DevEnv(Env):
    protocol = 'http://'
    host = 'dev0.avatar.vm.onti.actcognitive.org'


class TestEnv(Env):
    protocol = 'https://'
    host = 'avatar.dev-ops.zone/api/'


class ProdEnv(Env):
    protocol = 'https://'
    host = 'unknown'



import json
import os
from datetime import datetime
from logging import getLogger
from pathlib import Path
from time import sleep

import requests
from flask import Flask
from flask_sqlalchemy import SQLAlchemy

from config import app_path
from .models import CrawlerTask

logger = getLogger()


class BaseCrawlerAdapter():
    tasks_path: Path = app_path.joinpath('crawling', 'tasks')

    @property
    def crawler_create_task_url(self):
        return f'{self.crawler_url}/task/create'

    @property
    def crawler_run_task_url(self):
        return f'{self.crawler_url}/task/run'

    @property
    def crawler_task_status_url(self):
        return f'{self.crawler_url}/task/status'

    @property
    def crawler_data_url(self):
        return f'{self.crawler_url}/data/get'

    def __init__(self, app: Flask, db: SQLAlchemy, tasks_path: Path = app_path.joinpath('crawling', 'tasks')):
        self.app = app
        self.db = db
        self.tasks_path = tasks_path

        self.crawler_url = app.config.get('CRAWLER_URL')

        logger.debug(f'CrawlerAdapter created')

        if app.config.get('CRAWLER_ADD_TASKS'):
            self.__add_tasks()
            logger.debug(f'Tasks from {self.tasks_path} added')

    def __add_tasks(self):
        logger.debug(f'Scan {self.tasks_path} directory')

        with os.scandir(self.tasks_path) as entries:
            entry: os.DirEntry
            for entry in entries:
                with open(entry, mode='r', encoding='utf8') as file:
                    task_body = json.load(file)
                    task_name = task_body.get('name')
                    if not task_name:
                        logger.error(f'Cannot import task from file {file.name}. Specify task name in task.name')
                        continue

                    self.__create_crawler_task(task_name=task_name, task_body=task_body)

    def __create_crawler_task(self, task_name: str, task_body: str):
        response = requests.post(self.crawler_create_task_url, json=task_body)

        if response.status_code != 200:
            logger.warning(f'Response status is {response.status_code}: Payload: {response.text}')

        data = response.json()
        if data:
            task_id = data['result']['id']

            crawler_task: CrawlerTask = CrawlerTask.query.filter_by(name=task_name).first()
            if not crawler_task:
                crawler_task = CrawlerTask()
                crawler_task.name = task_name
                self.db.session.add(crawler_task)

            crawler_task.task_id = task_id
            crawler_task.task_body = json.dumps(task_body)
            self.db.session.commit()

            logger.debug(f'Task {task_name} added')
        else:
            logger.warning('Response data is empty')

    def __run_task(self, crawler_task: CrawlerTask):
        task = json.loads(crawler_task.task_body)
        task['task_id'] = crawler_task.task_id

        response = requests.post(self.crawler_run_task_url, json=task)

        if response.status_code != 200:
            logger.error(f'Run task is not succeed. Status {response.status_code}. Payload: {response.text}')

        logger.debug(response.text)

    def __wait_until_task_finish(self, crawler_task: CrawlerTask, timeout: int):
        data = {
            'task_id': crawler_task.task_id
        }

        while True:
            response = requests.post(self.crawler_task_status_url, json=data)

            if response.status_code != 200:
                logger.warning(f'Check task status is not succeed. Status {response.status_code}. Payload: {response.text}')
            logger.debug(response.text)

            status = response.json()['result']['status']
            logger.debug(f'{crawler_task.name} task status: {status}')
            if status == 'in_db':
                break

            sleep(timeout)

    def __fetch(self, crawler_task: CrawlerTask, limit):
        task = {
            'task_id': crawler_task.task_id,
            'start_ts': crawler_task.last_crawled
        }
        if limit:
            task['limit'] = limit

        response = requests.post(self.crawler_data_url, json=task)

        data = response.json()['result']['data']

        logger.info(f'Crawled {len(data)} pieces of data')
        ctimes = [item.get('ctime') for item in data if item.get('ctime')]
        if ctimes:
            last_crawled = max(ctimes)
            logger.debug(f'last_crawled is {str(datetime.fromtimestamp(last_crawled))}')
            if last_crawled:
                crawler_task.last_crawled = last_crawled

        self.db.session.add(crawler_task)
        self.db.session.commit()

        return data

    def data(self, task_name: str, limit=None, timeout: int = 15):
        crawler_task: CrawlerTask = CrawlerTask.query.filter_by(name=task_name).first()
        if not crawler_task:
            raise KeyError(f'task {task_name} does not exist')

        if self.app.config.get('CRAWLER_RUN_TASK'):
            self.__run_task(crawler_task=crawler_task)

        if self.app.config.get('CRAWLER_WAIT_UNTIL_TASK_FINISH'):
            self.__wait_until_task_finish(crawler_task=crawler_task, timeout=timeout)

        data = self.__fetch(crawler_task=crawler_task, limit=limit)

        return data



from dataclasses import dataclass as python_dataclass, field
from typing import Any, List

from marshmallow_dataclass import dataclass

from avatar_utils.objects.abstracts.serializable import Serializable


@dataclass
@python_dataclass
class AbstractQuestionWithOptions(Serializable):

    def retrieve_answers(self):
        answers = []
        for answer_id in self.answers:
            answers.append(self.options[answer_id])

        return answers


@dataclass
@python_dataclass
class ChoiseQuestion(AbstractQuestionWithOptions):
    options: List[Any] = field(default_factory=list)
    answers: List[Any] = field(default_factory=list)
    label: str = None

    repr_type: str = None

    id: int = None


@dataclass
@python_dataclass
class SelectQuestion(AbstractQuestionWithOptions):
    options: List[Any] = field(default_factory=list)
    answers: List[Any] = field(default_factory=list)
    label: str = None

    repr_type: str = None

    id: int = None


@dataclass
@python_dataclass
class CheckBoxQuestion(AbstractQuestionWithOptions):
    options: List[Any] = field(default_factory=list)
    answers: List[Any] = field(default_factory=list)
    label: str = None

    repr_type: str = None

    id: int = None


@dataclass
@python_dataclass
class RangeQuestion(Serializable):
    answers: List[Any] = field(default_factory=list)
    min_value: float = None
    max_value: float = None
    label: str = None

    repr_type: str = None

    id: int = None


@dataclass
@python_dataclass
class TextFieldQuestion(Serializable):
    answers: List[Any] = field(default_factory=list)
    shadow_title: str = None
    validation_type: str = None
    label: str = None

    repr_type: str = None

    id: int = None



AUTH_HEADER_NAME = 'Authorization'
DEFAULT_TOKEN_TYPE = 'Bearer'
ITMO_REALM = 'itmo'
ITMO_SSO_SERVER = 'https://sso.itmo.ru/auth/'
PUBLIC_KEY_WRAPPER = '-----BEGIN PUBLIC KEY-----\n{public_key}\n-----END PUBLIC KEY-----'
USER_TOKEN_AUDIENCE = 'account'



from typing import Optional, Union

from flasgger import Swagger
from flask import Flask


def create_swagger(app: Flask,
                   title: str,
                   description: str,
                   version: Optional[Union[str, int]] = None,
                   host: Optional[str] = None,
                   show_ui: bool = True,
                   ui_route: str = '/api',
                   spec_route: str = '/api.json') -> Swagger:

    if version and not isinstance(version, str):
        version = str(version)

    template = _make_template(title=title, description=description, version=version, host=host)

    swagger_config: dict = Swagger.DEFAULT_CONFIG.copy()

    swagger_config['swagger_ui'] = show_ui
    swagger_config['specs_route'] = ui_route
    swagger_config['specs'][0]['route'] = spec_route

    swagger = Swagger(app, config=swagger_config, template=template)

    return swagger


def _make_template(title: str,
                   description: str,
                   version: Optional[str] = None,
                   host: Optional[str] = None,
                   **kwargs) -> dict:
    template = {
        "swagger": "2.0",
        "info": {
            "title": title,
            "description": description,
            "version": version
        },
        "schemes": [
            "http",
            "https"
        ]
    }
    if host:
        template.update({'host': host})

    template.update(kwargs)

    return template



from dataclasses import dataclass as python_dataclass

from marshmallow_dataclass import dataclass

from avatar_utils.objects.abstracts.abstract_object import AbstractObject
from avatar_utils.objects.person import Person


@dataclass
@python_dataclass
class Match(AbstractObject):
    person1: Person = None
    person2: Person = None

    repr_type: str = None

    score: float = None

    id: int = None
    stars: int = None



from dataclasses import dataclass as python_dataclass, field
from typing import Any, List

from marshmallow_dataclass import dataclass

from avatar_utils.objects.abstracts.abstract_object import AbstractObject


@dataclass
@python_dataclass
class Event(AbstractObject):
    title: str = None
    text: str = None
    origin: str = None
    deadline: str = None
    place: str = None
    tags: List[Any] = None

    repr_type: str = None

    id: int = None
    stars: int = None



from time import sleep

from flask_sqlalchemy import BaseQuery
from sqlalchemy.exc import OperationalError


class RetryingQuery(BaseQuery):
    __retry_count__ = 3
    __retry_sleep_interval_sec__ = 0.5

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __iter__(self):
        attempts = 0
        while True:
            attempts += 1
            try:
                return super().__iter__()
            except OperationalError as ex:
                if "server closed the connection unexpectedly" not in str(ex):
                    raise
                if attempts < self.__retry_count__:
                    print("Connection lost - sleeping for %.2f sec and will retry (attempt #%d)",
                          self.__retry_sleep_interval_sec__, attempts)
                    sleep(self.__retry_sleep_interval_sec__)
                    continue
                else:
                    raise



import asyncio
from logging import getLogger
from time import time
from typing import Optional, List
from uuid import uuid4

import aiohttp
from aiohttp import ClientSession, ClientConnectorError

from .request_data import RequestData
from .requests_result import RequestsResult
from .response_data import ResponseData

logger = getLogger()
ASYNCIO_SEMAPHORE = 1024


async def async_make_requests(*requests_data) -> RequestsResult:
    call_id = str(uuid4()).replace('-', '')

    requests_data: List[RequestData] = [RequestData.make(req) for req in requests_data]
    semaphore = asyncio.Semaphore(ASYNCIO_SEMAPHORE)
    results = list()
    tasks = list()
    start_at = time()

    requests_data_expanded = list()

    for req in requests_data:
        for i in range(req.count):
            requests_data_expanded.append(dict(method=req.method,
                                               url=req.url,
                                               json=req.json,
                                               headers=req.headers,
                                               timeout=req.timeout))

    requests_count = len(requests_data_expanded)
    requests_id_len = len(str(requests_count))

    async with ClientSession() as session:
        for i in range(requests_count):
            method = requests_data_expanded[i]['method']
            url = requests_data_expanded[i]['url']
            request_id = f'{call_id} | {i+1:0>{requests_id_len}}'
            task = asyncio.create_task(_request_fetch(session=session,
                                                      semaphore=semaphore,
                                                      method=method,
                                                      url=url,
                                                      headers=requests_data_expanded[i]['headers'],
                                                      json=requests_data_expanded[i]['json'],
                                                      request_id=request_id))
            # bodies_size += getsizeof(dumps(req.json)) if req.json is not None else 0
            tasks.append(task)
        await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
        # await asyncio.gather(*tasks)
        results = [task.result() for task in tasks]

    result = RequestsResult(responses=results,
                            elapsed=time()-start_at)
    logger.info(f'[ {call_id} ] ASYNC REQUESTS RESULT:\n'
                f'  Requests count: {len(result.responses)}\n'
                f'  Elapsed time:   {result.elapsed} seconds\n'
                f'  Average time:   {result.average} seconds\n'
                f'  Statuses:       {result.status_codes}\n'
                f'  Failed:         {result.failed}')

    return result


async def _request_fetch(session: ClientSession,
                         semaphore: asyncio.Semaphore,
                         method: str,
                         url: str,
                         headers: Optional[dict] = None,
                         json: Optional[dict] = None,
                         timeout=None,
                         request_id: Optional[str] = None) -> Optional[ResponseData]:

    async with semaphore:
        started_at = time()
        logger.info(f'[ {request_id} ] -> start task with {method} request to {url}')
        # headers = {'Content-Type': 'application/json'}  if json else None

        try:
            async with session.request(
                    method=method,
                    url=url,
                    json=json,
                    headers=headers,
                    ssl=False,
                    timeout=timeout,
            ) as response:
                await response.read()

        except asyncio.TimeoutError as err:
            logger.warning(f'[ {request_id} ] {err.__class__.__name__} {url}')
            return ResponseData(exception=err, url=url, started_at=started_at)
        except ClientConnectorError as err:
            logger.warning(f'[ {request_id} ] {err.__class__.__name__} {url}')
            return ResponseData(exception=err, url=url, started_at=started_at)

    finished_at = time()
    logger.info(f'[ {request_id} ] <- receive response from {url} ({finished_at-started_at:.3f} sec.)')

    response_data = ResponseData(response=response,
                                 url=url,
                                 started_at=started_at,
                                 finished_at=finished_at)
    response_data.text = await response.text()
    response_data.json = None

    try:
        response_data.json = await response.json()
    except aiohttp.ContentTypeError:
        logger.warning(f'[ {request_id} ] is not JSON response ({response.status} {response.reason})')

    return response_data



import traceback
from functools import wraps
from logging import getLogger
from typing import Type, Optional, Union, Callable, Tuple, Sequence

from flask import request, jsonify
from marshmallow import Schema
from marshmallow import ValidationError
from sqlalchemy.exc import DatabaseError
from werkzeug import Response, exceptions  # noqa: PyPackageRequirements

from app import db
from avatar_utils.core import create_response
from avatar_utils.validation.constants import ServiceHTTPCodes

logger = getLogger(__name__)

DEFAULT_VALIDATE_METHOD = 'GET'


def validate_json(request_schema: Optional[Type[Schema]] = None,
                  response_schema: Optional[Type[Schema]] = None,
                  methods: Union[str, Sequence[str]] = DEFAULT_VALIDATE_METHOD) -> Callable:
    if isinstance(methods, str):
        methods = {methods}

    methods = set(m.upper().strip() for m in methods)

    def decorator(f) -> Callable:

        @wraps(f)
        def wrapper(*args, **kw) -> Union[str, Tuple[Response, int]]:

            log_prefix = f'[ VALIDATION | {request.remote_addr} {request.method} > {request.url_rule} ]'

            # skip decorator if unexpected HTTP method of request
            if request.method not in methods:
                logger.debug(f'{log_prefix} skip {request.method} method, validator expect {methods}')
                return f(*args, **kw)

            # validate request data
            if request_schema:
                validation_schema = request_schema()

                if not request.is_json:
                    logger.warning(f'{log_prefix} < invalid content-type header')
                    return create_response(status=ServiceHTTPCodes.BAD_REQUEST.value,
                                           message='invalid content-type header')
                try:
                    validation_schema.load(request.json)
                    logger.debug(f'{log_prefix} request body validated')
                except exceptions.BadRequest:
                    logger.warning(f'{log_prefix} < invalid requests body')
                    return create_response(status=ServiceHTTPCodes.BAD_REQUEST.value,
                                           message='invalid requests body')
                except ValidationError as err:
                    logger.warning(f'{log_prefix} < validation error')
                    return create_response(status=ServiceHTTPCodes.UNPROCESSABLE_ENTITY.value,
                                           message='validation error',
                                           data=dict(err.messages))

            # call HTTP handler
            try:
                status = 200
                result = f(*args, **kw)
                if isinstance(result, tuple):
                    response, status = result
                else:
                    response = result
                if not isinstance(response, Response):
                    response = jsonify(result)
            except NotImplementedError as err:
                response, status = create_response(status=ServiceHTTPCodes.NOT_IMPLEMENTED.value,
                                                   message=type(err).__name__,
                                                   data=dict(traceback=str(traceback.format_exc())))
            except DatabaseError as err:
                logger.error(f'{log_prefix} < Rollback transaction due to: {err}')
                db.session.rollback()
                raise
            except Exception as err:  # noqa
                logger.error(f'{log_prefix} < {err}')
                response, status = create_response(status=ServiceHTTPCodes.SERVER_ERROR.value,
                                                   message=type(err).__name__,
                                                   data=dict(traceback=str(traceback.format_exc())))
                return response, status

            # validate response data
            if response_schema:
                validation_schema = response_schema()

                if response.status_code != ServiceHTTPCodes.OK.value:
                    return response, status

                try:
                    data = validation_schema.load(response.json)
                    logger.debug(f'{log_prefix} response validated')
                    response = jsonify(validation_schema.dump(data))
                except ValidationError as err:
                    logger.error(f'{log_prefix} < {err}')
                    response, status = create_response(status=ServiceHTTPCodes.SERVER_ERROR.value,
                                                       message=type(err).__name__,
                                                       data=dict(errors=dict(err.messages),
                                                                 response=response.json,
                                                                 traceback=str(traceback.format_exc())))
                except (exceptions.BadRequest, ValidationError, TypeError) as err:
                    logger.error(f'{log_prefix} < {err}')
                    response, status = create_response(status=ServiceHTTPCodes.SERVER_ERROR.value,
                                                       message=type(err).__name__,
                                                       data=dict(traceback=str(traceback.format_exc())))
                finally:
                    return response, status

            return response, status

        return wrapper

    return decorator



import json
from datetime import datetime
from logging import getLogger

from flask import request

from app import db

logger = getLogger(__name__)


def format_text(text):
    return ' '.join(text.strip().split())


class Log(db.Model):
    __tablename__ = 'logs'
    id = db.Column(db.Integer(), primary_key=True)
    route = db.Column(db.String(100))
    method = db.Column(db.String(10))
    request = db.Column(db.Text())
    response = db.Column(db.Text())
    ts = db.Column(db.DateTime(), default=datetime.utcnow)
    duration = db.Column(db.Integer())
    status_code = db.Column(db.String)
    message = db.Column(db.String)

    def __repr__(self):
        return """<log object>
        id: {}
        route: {}
        method: {}
        request: {}
        response: {}
        ts: {}
        duration: {}
        status_code: {}
        message: {}""".format(self.id, self.route, self.method, self.request, self.response, self.ts,
                              self.duration, self.status_code, self.message)

    @staticmethod
    def init(request):
        data = request.json
        log = Log()
        log.route = request.path
        log.method = request.method
        log.request = json.dumps(data)
        db.session.add(log)
        db.session.commit()
        return log

    @staticmethod
    def complete(log, response):
        if log:
            log.duration = (datetime.utcnow() - log.ts).total_seconds()
            log.status_code = response.status
            try:
                data = response.json
                log.response = json.dumps(data)
                log.message = data.get('message')
            except TypeError as err:
                log_prefix = f'[ EVENT LOGGING | {request.remote_addr} {request.method} > {request.url_rule} ]'
                logger.warning(f'{log_prefix} < Response is not in JSON format')
                logger.warning(
                    f'{log_prefix} < Message: {err.__class__.__name__}. Errors: {list(str(e) for e in err.args)}')
            db.session.commit()
            return True
        return False

    def delete(self):
        log = Log.query.filter_by(id=self.id).first()
        db.session.delete(log)
        db.session.commit()



import copy
import logging
from unittest import TestCase

import flask
import requests

from avatar_utils.tests.Env import DevEnv

logger = logging.getLogger(__name__)


class AbstractApp:
    client = None
    headers = {'Content-type': 'application/json'}

    def post(self, url, data=dict()) -> flask.Response:
        return self.client.post(url, json=data, headers=self.headers)

    def get(self, url, data=dict()) -> flask.Response:
        return self.client.get(url, json=data, headers=self.headers)

    def register(self, username, email, password, confirm) -> flask.Response:
        return self.post('/register', data=dict(username=username, email=email, password=password, confirm=confirm))

    def login(self, username, password) -> flask.Response:
        return self.post('/login', data=dict(username=username, password=password))

    def logout(self) -> flask.Response:
        return self.post('/logout')

    def delete(self, user_id, password):
        return self.post('/user/delete', data=dict(user_id=user_id, password=password))

    def rprint(self, response):
        if not isinstance(response, flask.Response):
            logger.warning(f'rprint prints only flask.Response class data')
            return

        json = response.json
        if json:
            logger.debug(json)
        else:
            logger.debug(response.data.decode())


class RemoteApp(AbstractApp):
    client = requests

    # default env, specified once per whole test class
    default_env = DevEnv()

    # current env, could be changed in every test fuction
    current_env = copy.deepcopy(default_env)

    @property
    def base_url(self):
        return self.current_env.base_url

    def post(self, url, data=dict()) -> flask.Response:
        res = super().post(self.base_url + url, data)
        return self.convert_to_flask_response(res)

    def get(self, url, data=dict()) -> flask.Response:
        res = super().get(self.base_url + url, data)
        return self.convert_to_flask_response(res)

    # TODO transfer headers
    @staticmethod
    def convert_to_flask_response(response: requests.Response) -> flask.Response:
        res = flask.Response(
            response=response.text,
            status=response.status_code,
            mimetype=response.headers['content-type']
        )

        return res


class UnitTest(TestCase, AbstractApp):
    app = None
    db = None

    def setUp(self):
        TestCase.setUp(self)
        try:
            print('SQLALCHEMY_DATABASE_URI: %s' % self.app.config['SQLALCHEMY_DATABASE_URI'])
        except KeyError:
            pass

        self.assertFalse(self.app is None)
        self.assertFalse(self.client is None)
        self.assertTrue(self.app.config['TESTING'])

        self.app_context = self.app.app_context()
        self.app_context.push()
        self.db.create_all()
        super().setUp()

    def tearDown(self):
        self.db.session.remove()
        self.db.drop_all()
        self.app_context.pop()
        TestCase.tearDown(self)


class BaseUnitTest(UnitTest):
    from app import create_app, db

    app = create_app('testing')
    client = app.test_client()
    db = db

    app.logger.handlers.clear()
    logging.basicConfig(level=app.config['LOGGING_LEVEL'],
                        format='%(asctime)s %(levelname)s %(message)s')


class IntegrationTest(TestCase, RemoteApp):

    def setUp(self):
        super().setUp()

        # reset current env
        self.current_env = copy.deepcopy(RemoteApp.default_env)



from logging import getLogger

from flask import g, request
from sqlalchemy.exc import DatabaseError

from . import logs
from .models import *

logger = getLogger(__name__)


@logs.before_app_request
def before_request():
    log_prefix = f'[ LOGGING BEFORE | {request.remote_addr} {request.method} > {request.url_rule} ]'
    try:
        log = Log.init(request)
        g.log = log
    except DatabaseError as err:
        logger.warning(f'{log_prefix} < Rollback transaction due to: {err}')
        db.session.rollback()
    except BaseException as err:
        logger.error(f'{log_prefix} < Error occurs: {err}')


@logs.after_app_request
def after_request(response):
    log_prefix = f'[ LOGGING AFTER | {request.remote_addr} {request.method} > {request.url_rule} ]'
    try:
        if g.log:
            Log.complete(g.log, response)
        else:
            logger.warning('log_id is None, cannot complete logging process')
    except DatabaseError as err:
        logger.warning(f'{log_prefix} < Rollback transaction due to: {err}')
        db.session.rollback()
    except BaseException as err:
        logger.error(f'{log_prefix} < Error occurs: {err}')
    return response



from enum import IntEnum, Enum, unique, auto


@unique
class ServiceHTTPCodes(IntEnum):
    OK = 200
    BAD_REQUEST = 400
    UNPROCESSABLE_ENTITY = 422
    SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501



import json
from dataclasses import dataclass as python_dataclass
from typing import Any


@python_dataclass
class Serializable:
    # import avatar_utils.objects - required
    import avatar_utils.objects

    def __post_init__(self):
        self.repr_type = self.__fullname()

    # reliable but not the most productive way to serialize as a dict
    def to_dict(self):
        json_data = json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

        return json.loads(json_data)

    @staticmethod
    def from_dict(data: Any):
        # import avatar_utils.objects - required
        import avatar_utils.objects

        if isinstance(data, dict):
            repr_type = data.pop('repr_type', None)

            # serializable class
            if repr_type:
                cls = eval(repr_type)

                result = {}
                for k, v in data.items():
                    result[k] = Serializable.from_dict(v)

                extracted_cls = cls(**result)
                return extracted_cls
            return data
        elif isinstance(data, list):
            i: int
            for i in range(data.__len__()):
                data[i] = Serializable.from_dict(data[i])
            return data
        else:
            return data

    def __fullname(self):
        # o.__module__ + "." + o.__class__.__qualname__ is an example in
        # this context of H.L. Mencken's "neat, plausible, and wrong."
        # Python makes no guarantees as to whether the __module__ special
        # attribute is defined, so we take a more circumspect approach.
        # Alas, the module name is explicitly excluded from __qualname__
        # in Python 3.

        module = self.__class__.__module__
        if module is None or module == str.__class__.__module__:
            return self.__class__.__name__  # Avoid reporting __builtin__
        else:
            return module + '.' + self.__class__.__name__



from .async_requests import async_make_requests
from .request_data import RequestData
from .requests_result import RequestsResult
from .response_data import ResponseData
from .sync_requests import make_requests

__all__ = ['RequestData',
           'ResponseData',
           'RequestsResult',
           'make_requests',
           'async_make_requests']



from logging import getLogger
from typing import Dict
from typing import Tuple

import requests
from flask import jsonify
from flask.wrappers import Response

from avatar_utils.objects.abstracts.serializable import Serializable

logger = getLogger(__name__)


class RequestTool:
    def post(self, url, data=dict()):
        headers = {'Content-type': 'application/json'}
        return requests.post(url, json=data, headers=headers)


class ResponseTool:

    def __init__(self, default_headers=dict()):
        self.default_headers = default_headers

    def create(self, data: dict = None, status: int = 200, message: str = '', headers=dict()) -> \
            Tuple[Response, int]:
        """Wraps response in a consistent format throughout the API.

        Format inspired by https://medium.com/@shazow/how-i-design-json-api-responses-71900f00f2db
        Modifications included:
        - make success a boolean since there's only 2 values
        - make message a single string since we will only use one message per response

        IMPORTANT: data must be a dictionary where:
        - the key is the name of the type of data
        - the value is the data itself

        :param headers <dict> Dictionary of headers
        :param data <dict> optional data
        :param status <int> optional status code, defaults to 200
        :param message <str> optional message
        :returns tuple of Flask Response and int
        """

        if data is not None:
            if not isinstance(data, Dict):
                try:
                    if isinstance(data, Serializable):
                        data = data.to_dict()
                    else:
                        data = data.__dict__
                except AttributeError as err:
                    logger.error('Cannot cast object to dict representation. Raise error.')
                    raise TypeError(f'Data should be a dictionary 😞. {err}')

        # merge default headers with current
        headers = {**self.default_headers, **headers}
        res = {'success': 200 <= status < 300, 'message': message, 'result': data,
               'headers': headers}
        response = jsonify(res)
        response.status_code = status
        return response, status

    def copy(self, response, message=None):
        if response is not None:
            return self.create(data=response.json()['result'], status=response.status_code,
                               message=message if message else response.json()['message'])
        else:
            raise TypeError('Response should not be None')


req = RequestTool()
post = req.post

resp = ResponseTool()
create_response = resp.create
copy_response = resp.copy



from dataclasses import dataclass
from typing import Optional, Sequence

from aiohttp import ClientResponse


@dataclass
class ResponseData:

    url: str

    started_at: float
    elapsed: float = None

    request_schema: str = None
    host: str = None
    port: int = None
    route: str = None
    route_params: str = None

    finished_at: Optional[float] = None

    text: Optional[str] = None
    json: Sequence = None

    response: Optional[ClientResponse] = None
    exception: Optional[Exception] = None

    def __post_init__(self):
        if self.finished_at:
            self.elapsed = self.finished_at - self.started_at

        self.request_schema, self.host = self.url.split('//', 1)
        self.port = 443 if self.request_schema.lower() == 'https' else 80

        if '/' in self.host:
            self.host, self.route = self.host.split('/', 1)

        self.route = '/' + self.route

        if ':' in self.host:
            self.host, self.port = self.host.split(':', 1)

        if '?' in self.route:
            self.route, self.route_params = self.host.split('?', 1)



import atexit
from logging import getLogger
from typing import Any

from apscheduler.schedulers.background import BackgroundScheduler
from flask import Flask

logger = getLogger()


class BaseLearnScheduler:

    def __init__(self, app: Flask, learn_func: Any, learn_args: Any = ()):
        self.app = app
        self.learn_func = learn_func
        self.learn_args = learn_args

        if app.config.get('LEARN_SCHEDULE'):
            self.schedule_learn()
        if app.config.get('LEARN_RUN_AT_START'):
            self.run_learn()

    def cron_trigger_args(self):
        trigger_args = {'day': self.app.config.get('LEARN_DAY'),
                        'day_of_week': self.app.config.get('LEARN_DAY_OF_WEEK'),
                        'hour': self.app.config.get('LEARN_HOUR'),
                        'minute': self.app.config.get('LEARN_MINUTE')}
        return dict((arg, value) for arg, value in trigger_args.items() if value is not None)

    def interval_trigger_args(self):
        trigger_args = {'days': self.app.config.get('LEARN_DAYS'),
                        'hours': self.app.config.get('LEARN_HOURS'),
                        'minutes': self.app.config.get('LEARN_MINUTES'),
                        'seconds': self.app.config.get('LEARN_SECONDS'),
                        'start_date': self.app.config.get('LEARN_START_DATE'),
                        'end_date': self.app.config.get('LEARN_END_DATE')}
        return dict((arg, value) for arg, value in trigger_args.items() if value is not None)

    def schedule_learn(self):
        scheduler = BackgroundScheduler()
        args_dict = {
            'cron': self.cron_trigger_args,
            'interval': self.interval_trigger_args
        }

        try:
            trigger = self.app.config.get('LEARN_TRIGGER')
            trigger_args = args_dict[trigger]()
        except ValueError as err:
            logger.error(err)
            return

        scheduler.add_job(func=self.learn_func, args=self.learn_args, trigger=trigger, **trigger_args)

        scheduler.start()
        atexit.register(lambda: scheduler.shutdown())
        logger.debug(f'Learn scheduled')

    def run_learn(self):
        self.learn_func(*self.learn_args)



from typing import Optional, Tuple, Dict

from flask import request

from avatar_utils.sso_helper.constants import (
    AUTH_HEADER_NAME,
    DEFAULT_TOKEN_TYPE,
)


class FlaskAuthHeader:

    @property
    def from_request(self) -> Optional[str]:

        return request.headers.get(AUTH_HEADER_NAME)

    @property
    def request_token(self) -> Optional[str]:

        value, _ = self.extract()
        return value

    @property
    def request_token_type(self) -> Optional[str]:

        _, value = self.extract()
        return value

    def extract(self) -> Tuple[Optional[str], Optional[str]]:

        raw_value = self.from_request
        token_data = raw_value.split(' ') if raw_value else (None, None)
        token_type, token_string = token_data
        return token_string, token_type

    def make_header_value(self,
                          token_string: Optional[str] = None,
                          token_type: Optional[str] = None) -> Optional[str]:

        if not token_type and token_string and token_string.startswith(DEFAULT_TOKEN_TYPE):
            token_type, token_string = token_string.split(' ')

        if not token_type and token_string:
            token_type = DEFAULT_TOKEN_TYPE

        if not token_type and not token_string:
            token_string, token_type = self.extract()

        return f'{token_type} {token_string}'

    def make_header_dict(self,
                         token_string: Optional[str] = None,
                         token_type: Optional[str] = None) -> Dict[str, Optional[str]]:

        value = self.make_header_value(token_string, token_type)
        return {AUTH_HEADER_NAME: value}



from dataclasses import dataclass as python_dataclass, field
from typing import List

from marshmallow_dataclass import dataclass

from avatar_utils.objects.abstracts.abstract_object import AbstractObject


@dataclass
@python_dataclass
class Article(AbstractObject):
    author_list: List[str] = field(default_factory=list)
    title: str = None
    journal_info: str = None
    source: str = None

    repr_type: str = None

    id: int = None
    stars: int = None




import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='Avatar-Utils',
    version='1.3.6',
    author='Algorithmics of Complex System',
    author_email='artem.sementsov@gmail.com',
    description='Common utils for services in ecosystem',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='',
    packages=[
        'avatar_utils.async_requests',
        'avatar_utils.core',
        'avatar_utils.crawler',
        'avatar_utils.db',
        'avatar_utils.learn_scheduler',
        'avatar_utils.sso_helper',
        'avatar_utils.logs',
        'avatar_utils.registration',
        'avatar_utils.tests',
        'avatar_utils.validation',
        'avatar_utils.objects',
        'avatar_utils.objects.abstracts',
    ],
    install_requires=[
        'aiohttp>=3.6.2',
        'flask>=1.1.1',
        'flask-sqlalchemy>=2.3.2',
        'psycopg2-binary>=2.7.5',
        'requests>=2.23.0',    # registration
        'apispec>=3.3.0',      # registration
        'marshmallow>=3.6.0',  # registration
        'flasgger>=0.9.4',     # registration
        'APScheduler~=3.6.3',
        'python-keycloak~=0.21.0',  # sso_helper
        'cryptography~=3.0',        # sso_helper
        'PyJWT~=1.7.1',             # sso_helper
        'marshmallow_dataclass~=7.6.0',  # objects
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
)



from dataclasses import dataclass, field
from typing import List, Dict, Optional

from avatar_utils.async_requests.response_data import ResponseData


@dataclass
class RequestsResult:

    responses: List[Optional[ResponseData]]
    elapsed: float

    average: float = field(default=None, init=False)
    min: float = field(default=None, init=False)
    max: float = field(default=None, init=False)
    status_codes: Dict[int, int] = field(default=None, init=False)
    failed: int = field(default=None, init=False)

    def __post_init__(self):
        self.average = self.elapsed / len(self.responses) if self.responses else None
        self.status_codes = dict()
        self.failed = 0
        elapsed_times = {r.elapsed for r in self.responses if r.elapsed is not None}

        if elapsed_times:
            self.min = min(elapsed_times)
            self.max = max(elapsed_times)

        for response in self.responses:
            if response.exception:
                self.failed += 1
                continue

            status_stat = self.status_codes.get(response.response.status)
            self.status_codes[response.response.status] = status_stat + 1 if status_stat else 1



import asyncio

from .async_requests import async_make_requests
from .requests_result import RequestsResult


def make_requests(*requests_data) -> RequestsResult:
    return asyncio.run(async_make_requests(*requests_data))



from avatar_utils.objects.article import Article
from avatar_utils.objects.event import Event
from avatar_utils.objects.feed_item import FeedItem
from avatar_utils.objects.image import Image
from avatar_utils.objects.interest import Interest
from avatar_utils.objects.match import Match
from avatar_utils.objects.message import Message
from avatar_utils.objects.person import Person
from avatar_utils.objects.question import *
